import base64
import hashlib
import os
import webbrowser
from multiprocessing import Queue

import requests
import uvicorn
from fastapi import APIRouter, FastAPI
from fastapi.responses import RedirectResponse

from oqtant.settings import Settings
from oqtant.util.server import ThreadServer

settings = Settings()

app = FastAPI(title="Login API", openapi_url="/openapi.json")
router = APIRouter()


def generate_random(length: int) -> str:
    return base64.urlsafe_b64encode(os.urandom(length)).decode("utf-8").replace("=", "")


verifier = generate_random(80)


def generate_challenge(verifier: str) -> str:
    hashed = hashlib.sha256(verifier.encode("utf-8")).digest()
    return base64.urlsafe_b64encode(hashed).decode("utf-8").replace("=", "")


def get_authentication_url():
    code_challenge = generate_challenge(verifier)
    auth_url = "".join(
        [
            f"{settings.auth0_base_url}/authorize",
            "?response_type=code",
            f"&scope={settings.auth0_scope}",
            f"&audience={settings.auth0_audience}",
            f"&code_challenge={code_challenge}",
            "&code_challenge_method=S256",
            f"&client_id={settings.auth0_client_id}",
            "&redirect_uri=http://localhost:8080",
        ]
    )
    return auth_url


queue = Queue()


@app.get("/")
async def main(code):
    resp = await get_token(verifier, code)
    token = resp["access_token"]
    queue.put({"token": token})
    if token:
        return "Successfully authenticated, you may close this tab now"
    else:
        return "Failed to authenticate, please close this tab and try again"


@app.get("/login")
def login():
    return RedirectResponse(url=get_authentication_url())


async def get_token(verifier: str, code: str):
    url = f"{settings.auth0_base_url}/oauth/token"
    headers = {"content-type": "application/x-www-form-urlencoded"}
    data = {
        "grant_type": "authorization_code",
        "client_id": settings.auth0_client_id,
        "code_verifier": verifier,
        "code": code,
        "redirect_uri": "http://localhost:8080",
    }
    resp = requests.post(
        url, headers=headers, data=data, allow_redirects=False, timeout=(5, 30)
    )
    return resp.json()


def get_user_token() -> str:
    """A utility function required for getting Oqtant authenticated with your Albert account.
       Starts up a server to handle the Auth0 authentication process and acquire a token.
    Returns:
        str: Auth0 user token
    """
    server_config = uvicorn.Config(
        app=app, host="localhost", port=8080, log_level="error"
    )
    server = ThreadServer(config=server_config)
    with server.run_in_thread():
        webbrowser.open("http://localhost:8080/login")
        token = queue.get(block=True)
    return token["token"]



from pydantic import BaseSettings

# class Settings(BaseSettings):
#     auth0_base_url: str = "https://coldquanta.auth0.com"
#     auth0_client_id: str = "cTWqizVsJj2WrLtBgzgo3KZFgEY3k2OC"
#     auth0_scope: str = "offline_access bec_service:client"
#     auth0_audience: str = "https://albert.coldquanta.com/oqtant"
#     signin_local_callback_url: str = "http://localhost:8080"
#     base_url: str = "https://albert.coldquanta.com/api/v1/bert/jobs"
#     max_ind_var: int = 2
#     max_job_batch_size: int = 30


# dev
class Settings(BaseSettings):
    auth0_base_url: str = "https://coldquanta-dev.us.auth0.com"
    auth0_client_id: str = "ZzQdn5ZZq1dmpP5N55KINr33u47RBRiu"
    auth0_scope: str = "offline_access bec_dev_service:client"
    auth0_audience: str = "https://albert-dev.coldquanta.com/oqtant"
    signin_local_callback_url: str = "http://localhost:8080"
    base_url: str = "https://albert-dev.coldquanta.com/api/v1/bert/jobs"
    max_ind_var: int = 2
    max_job_batch_size: int = 30



from datetime import datetime
from math import floor, log10, pi
from uuid import UUID

import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize as opt
from bert_schemas import job as job_schema
from matplotlib.ticker import FormatStrFormatter, LinearLocator
from pydantic import Field

from oqtant.util.exceptions import (
    JobPlotFitError,
    JobPlotFitMismatchError,
    JobReadError,
)

OMEGA_X = 2 * pi * 80
OMEGA_Y = 2 * pi * 1000


def round_sig(x: float, sig: int = 2):
    return round(x, sig - int(floor(log10(abs(x)))) - 1)


def TF_dist_2D(xy_mesh, TFpOD, xc, yc, rx, ry, os):
    """
    Defines 2D Thomas-Fermi distribution characteristic of zero-temperature Bose-gas
    Requires function(s): get_image_space
    :param xy_mesh:
    :type xy_mesh: (2,N,M) Matrix of floats containing mesh grid of image coordinates
    :param TFpOD:
    :type TFpOD: float - Thomas-Fermi peak Optical Density (OD)
    :param rx:
    :type rx: float - Thomas-Fermi radius along the x-direction
    :param ry:
    :type ry: float - Thomas-Fermi radius along the y-direction (along gravity)
    :param xc:
    :type xc: float - Cloud center along the x-direction (along gravity)
    :param yc:
    :type yc: float - Cloud center along the y-direction
    :param os:
    :type os: float - Constant offset
    """

    # unpack 1D list into 2D x and y coords
    (x, y) = xy_mesh

    # Simplify Thomas-Fermi expression
    A = 1 - ((y - yc) / rx) ** 2 - ((x - xc) / ry) ** 2

    # make 2D Thomas-Fermi distribution
    OD = np.real(TFpOD * np.maximum(np.sign(A) * (np.abs(A)) ** (3 / 2), 0)) + os

    # flatten the 2D Gaussian down to 1D
    return OD.ravel()


def Gaussian_dist_2D(xy_mesh, GpOD, xc, yc, sigx, sigy, os):
    """
    Defines 2D gaussian distribution characteristic of a thermal ensemble of atoms
    Requires function(s): get_image_space
    :param xy_mesh:
    :type xy_mesh: (2,N,M) Matrix of floats containing meshgrid of image coordinates
    :param GpOD:
    :type GpOD: float - Gaussian peak Optical Density (OD)
    :param sigx:
    :type sigx: float - Gaussian spread along the x-direction
    :param sigy:
    :type sigy: float - Gaussian spread along the y-direction (along gravity)
    :param xc:
    :type xc: float - Cloud center along the x-direction (along gravity)
    :param yc:
    :type yc: float - Cloud center along the y-direction
    :param os:
    :type os: float - Constant offset
    """

    (x, y) = xy_mesh

    OD = (
        GpOD * np.exp(-0.5 * ((y - yc) / sigy) ** 2 - 0.5 * ((x - xc) / sigx) ** 2) + os
    )
    return OD.ravel()


def bimodal_dist_2D(xy_mesh, GpOD, sigx, sigy, TFpOD, rx, ry, xc, yc, os):
    """
    Defines 2D bimodal distribution characteristic of finite-temperature Bose-gas
    Requires functions: Gaussian_dist_2D, TF_dist_2D, get_image_space
    :param xy_mesh:
    :type xy_mesh: (2,N,M) Matrix of floats containing meshgrid of image coordinates
    :param GpOD:
    :type GpOD: float - Gaussian peak Optical Density (OD)
    :param sigx:
    :type sigx: float - Gaussian spread along the x-direction
    :param sigy:
    :type sigy: float - Gaussian spread along the y-direction (along gravity)
    :param TFpOD:
    :type TFpOD: float - Thomas-Fermi peak Optical Density (OD)
    :param rx:
    :type rx: float - Thomas-Fermi radius along the x-direction
    :param ry:
    :type ry: float - Thomas-Fermi radius along the y-direction (along gravity)
    :param xc:
    :type xc: float - Cloud center along the x-direction (along gravity)
    :param yc:
    :type yc: float - Cloud center along the y-direction
    :param os:
    :type os: float - Constant offset
    """

    return Gaussian_dist_2D(xy_mesh, GpOD, xc, yc, sigx, sigy, os) + TF_dist_2D(
        xy_mesh, TFpOD, xc, yc, rx, ry, os
    )


class AlbertJob(job_schema.JobCreate):
    external_id: UUID | None
    time_submit: datetime | None
    pix_cal: float = Field(8.71, const=True)
    sig_abs: float = Field(0.297, const=True)
    omega_x: float = Field(OMEGA_X, const=True)
    omega_y: float = Field(OMEGA_Y, const=True)

    class Config:
        validate_assignment = True

    def get_TOF(self):
        """
        Returns shaped TOF image if it exists
        :returns: reshaped pixels numpy array (100,100)

        """
        if self.status != job_schema.JobStatus.COMPLETE:
            raise JobReadError("Job results not available.")
        try:
            reshaped_pixels = np.array(
                self.inputs[0].output.values.tof_image.pixels
            ).reshape(100, 100)
        except Exception as exc:
            raise JobReadError("no TOF results") from exc
        return reshaped_pixels

    def get_IT(self):
        """
        Returns shaped IT image if it exists
        :returns: reshaped pixels numpy array (100,100)

        """
        if self.status != job_schema.JobStatus.COMPLETE:
            raise JobReadError("Job results not available.")

        try:
            reshaped_pixels = np.array(
                self.inputs[0].output.values.it_plot.pixels
            ).reshape(148, 512)
        except Exception as exc:
            raise JobReadError("no IT results") from exc
        return reshaped_pixels

    @staticmethod
    def get_image_space(datafile=np.zeros((100, 100)), centered="y"):
        """
        Returns meshgrid of image coordinates
        :param datafile:
        :type datafile: (N,M) Matrix of Optical Density (OD) Data
        """

        lx, ly = np.shape(datafile)
        x, y = np.arange(lx), np.arange(ly)

        if centered == "y":
            x = x - round(lx / 2)
            y = y - round(ly / 2)

        xy_mesh = np.meshgrid(x, y)

        return xy_mesh, lx, ly

    def fit_bimodal_data2D(self, xi=None, lb=None, ub=None):
        """
        Performs fit via trust region reflective algorithm.
        Requires functions: bimodal_dist_2D, Gaussian_dist_2D, TF_dist_2D, get_image_space
        For better fit performance, tune initial guess 'xi' and lower/upper bounds, 'lb' and 'ub'
        :param xy_mesh:
        :type xy_mesh: (2,N,M) Matrix containing meshgrid of image data coordinates
        :param data2D:
        :type data2D: (N,M) Matrix containing image data
        :param xi:
        :type xi: (1,9) List of fit parameter initial guesses
        :param lb:
        :type lb:  (1,9) List of fit parameter lower bounds
        :param ub:
        :type ub: (1,9) List of fit parameter upper bounds
        """
        xi = xi if xi else [0.25, 8, 8, 1, 4, 6, 0, 0, 0.02]
        lb = lb if lb else [0, 7, 7, 0, 2, 2, -20, -20, 0]
        ub = ub if ub else [2, 20, 20, 2, 20, 20, 20, 20, 1]

        TOF_data = self.get_TOF()
        xy_mesh, _, _ = self.get_image_space()  # TOF_data)

        (X, Y) = xy_mesh
        x = X[0]
        y = Y[:, 0]

        fit_params, cov_mat = opt.curve_fit(
            bimodal_dist_2D, xy_mesh, np.ravel(TOF_data), p0=xi, bounds=(lb, ub)
        )
        fit_residual = TOF_data - bimodal_dist_2D(xy_mesh, *fit_params).reshape(
            np.outer(x, y).shape
        )
        fit_Rsquared = 1 - np.var(fit_residual) / np.var(TOF_data)

        return fit_params, cov_mat, fit_residual, fit_Rsquared

    def plot_fit_results(
        self, fit_params, model="bimodal", file_name=None, plot_title=None
    ):
        """
        Plot the results of a fit operation

        :param fit_params:
        :type fit_params: list of parameters from a fit operation
        :param model:
        :type model: string "bimodal", "TF", or "gaussian". default "bimodal"
        :param output:
        :type output: valid filename
        :param plot_title:
        :type plot_title: string title for the plot.
            default "job: "+str(self.name)+"\nTOF fit: "+str(model)

        """

        xy_mesh, _, _ = self.get_image_space()  # TOF_data)

        (X, Y) = xy_mesh

        if model == "bimodal":
            try:
                m = bimodal_dist_2D(xy_mesh, *fit_params)
            except TypeError as exc:
                raise JobPlotFitMismatchError() from exc
            except Exception as exc:
                raise JobPlotFitError() from exc

        elif model == "gaussian":
            try:
                m = Gaussian_dist_2D(xy_mesh, *fit_params)
            except TypeError as exc:
                raise TypeError(
                    "PLOT FIT RESULTS: mismatched parameters and model type"
                ) from exc
            except Exception as exc:
                raise JobPlotFitError() from exc
        elif model == "TF":
            try:
                m = TF_dist_2D(xy_mesh, *fit_params)
            except TypeError as exc:
                raise JobPlotFitMismatchError() from exc
            except Exception as exc:
                raise JobPlotFitError() from exc
        else:
            print(
                f"PLOT FIT RESULTS: Invalid model specified: {model}.",
                " Select 'bimodal', 'gaussian', or 'TF'",
            )
            return

        m = m.reshape(100, 100)
        plt.figure()
        plt.imshow(
            m,
            origin="lower",
            cmap="nipy_spectral",
            extent=[
                np.min(X) * self.pix_cal,
                np.max(X) * self.pix_cal,
                np.min(Y) * self.pix_cal,
                np.max(Y) * self.pix_cal,
            ],
        )

        if plot_title is None:
            plot_title = f"job: {self.name}\nTOF fit: {model}"

        plt.title(plot_title)

        if file_name:
            self._save_plot_file(plt, file_name)
        plt.show()

    @staticmethod
    def _save_plot_file(plot, file_name):
        file = f"{file_name}.png"
        try:
            plot.savefig(file)
            print(f"plot saved to file: {file}")
        except (FileNotFoundError, Exception):
            print(f"failed to save plot at {file}")

    def atoms_2dplot(self, file_name=None, figsize=(12, 12), gridon=False):
        """
        Generate a 2D plot of atom OD (save or show)

        :param output: how to output the information
        :type output: string "show" or valid filename
        :param figsize:
        :type figsize: tuple. default is (12,12)
        :param gridon: grid lines on plot on/off
        :type gridon: Boolean. default is False

        """

        TOF_data = self.get_TOF()
        xy_mesh, _, _ = self.get_image_space()  # TOF_data)
        (X, Y) = xy_mesh

        fig2D = plt.figure(figsize=figsize)
        ax = fig2D.gca()
        plt2D = plt.imshow(
            TOF_data,
            origin="lower",
            cmap="nipy_spectral",
            extent=[
                np.min(X) * self.pix_cal,
                np.max(X) * self.pix_cal,
                np.min(Y) * self.pix_cal,
                np.max(Y) * self.pix_cal,
            ],
        )
        plt.grid(b=gridon)
        plt.colorbar(plt2D, shrink=0.8)

        ax.set_xlabel("x-position, micron", labelpad=15, fontsize=16)
        ax.set_ylabel("y-position, micron", labelpad=15, fontsize=16)
        plt.title("Time of Flight Optical Depth", fontsize=16)

        print("Peak OD: ", np.max(TOF_data))

        if file_name:
            self._save_plot_file(plt, file_name)
        plt.show()

    def atoms_sliceplot(self, file_name=None, sliceaxis="x", gridon=False):
        """
        Generate a 1D slice plot of atom OD in x or y

        :param output: how to output the information
        :type output: string "show" or valid filename
        :param sliceaxis:
        :type sliceaxis: string 'x' or 'y'
        :param figsize:
        :type figsize: tuple. default is (12,12)
        :param gridon: grid lines on plot on/off
        :type gridon: Boolean. default is False

        """
        TOF_data = self.get_TOF()
        xy_mesh, lx, ly = self.get_image_space(TOF_data)
        (X, Y) = xy_mesh

        params, *_ = self.fit_bimodal_data2D()
        fitOD = bimodal_dist_2D(xy_mesh, *params)

        Gfit_params = [params[0], params[6], params[7], params[1], params[2], params[8]]
        fitODG = Gaussian_dist_2D(xy_mesh, *Gfit_params)

        # Reshape Fit Distributions to 2D form
        fitOD2D = fitOD.reshape(lx, ly)
        fitODG2D = fitODG.reshape(lx, ly)

        # Define Central slices
        xslice = fitOD2D[int(lx / 2), :]
        yslice = fitOD2D[:, int(ly / 2)]
        xsliceG = fitODG2D[int(lx / 2), :]
        ysliceG = fitODG2D[:, int(ly / 2)]

        if sliceaxis == "x":
            xsliceD = TOF_data[int(len(X[1]) / 2), :]
            xslice = fitOD2D[int(len(X[1]) / 2), :]
            xsliceG = fitODG2D[int(len(X[1]) / 2), :]
            plt.title("X-Slice", fontsize=16)
            plt.plot(X[1] * self.pix_cal, xsliceD, "ok")
            plt.plot(X[1] * self.pix_cal, xslice, "b")
            plt.plot(X[1] * self.pix_cal, xsliceG, "r")
        elif sliceaxis == "y":
            ysliceD = TOF_data[:, int(len(Y[1]) / 2)]
            yslice = fitOD2D[:, int(len(Y[1]) / 2)]
            ysliceG = fitODG2D[:, int(len(Y[1]) / 2)]
            plt.title("Y-Slice", fontsize=16)
            plt.plot(Y[:, 1] * self.pix_cal, ysliceD, "ok")
            plt.plot(Y[:, 1] * self.pix_cal, yslice, "b")
            plt.plot(Y[:, 1] * self.pix_cal, ysliceG, "r")
        else:
            raise ValueError("Input either x or y")

        plt.grid(b=gridon)
        plt.xlabel("x-position, micron", labelpad=15, fontsize=16)
        plt.ylabel("Optical Depth", labelpad=15, fontsize=16)

        if file_name:
            self._save_plot_file(plt, file_name)
        plt.show()

    # This function plots the optical depth as a 3D surface with projected density contours
    def atoms_3dplot(self, file_name=None, view_angle=-45, figsize=(10, 10)):
        """
        Generate a 3D slice plot of atom OD

        :param output: how to output the information
        :type output: string "show" or valid filename
        :param view_angle:
        :type view_angle: int (-180, 180). default -45
        :param figsize:
        :type figsize: tuple. default is (10,10)

        """

        fig3d = plt.figure(figsize=figsize)
        ax = fig3d.gca(projection="3d")

        ax.zaxis.set_major_locator(LinearLocator(10))
        ax.zaxis.set_major_formatter(FormatStrFormatter("%.02f"))
        ax.xaxis.pane.fill = False
        ax.yaxis.pane.fill = False
        ax.zaxis.pane.fill = False

        # Set axis labels
        ax.set_xlabel("x-position, micron", labelpad=10)
        ax.set_ylabel("y-position, micron", labelpad=10)
        ax.set_zlabel("Optical Depth", labelpad=10)

        # rotate the axes and update
        ax.view_init(30, view_angle)

        if file_name:
            self._save_plot_file(plt, file_name)
        plt.show()

    def atom_numbers(self, bimodalfit_params, print_results=True):
        xy_mesh, lx, ly = self.get_image_space()

        # lmfit Toolbox Results
        # Remove Background
        bimodalfit_params[8] = bimodalfit_params[8] * 0

        # Define Fit Distributions in 1D form
        fitOD = bimodal_dist_2D(xy_mesh, *bimodalfit_params)
        Gfit_params = [
            bimodalfit_params[0],
            bimodalfit_params[6],
            bimodalfit_params[7],
            bimodalfit_params[1],
            bimodalfit_params[2],
            bimodalfit_params[8],
        ]
        TFfit_params = [
            bimodalfit_params[3],
            bimodalfit_params[6],
            bimodalfit_params[7],
            bimodalfit_params[4],
            bimodalfit_params[5],
            bimodalfit_params[8],
        ]
        fitODG = Gaussian_dist_2D(xy_mesh, *Gfit_params)
        fitODTF = TF_dist_2D(xy_mesh, *TFfit_params)
        lmfitbg = lx * ly * bimodalfit_params[8]

        # Reshape Fit Distributions to 2D form
        fitOD2D = fitOD.reshape(lx, ly)
        fitODG2D = fitODG.reshape(lx, ly)
        fitODTF2D = fitODTF.reshape(lx, ly)

        ODsumtot = np.sum(fitOD2D) - lmfitbg * 2
        ODsumcond = np.sum(fitODTF2D) - lmfitbg
        ODsumtherm = np.sum(fitODG2D) - lmfitbg
        Ncond = self.pix_cal**2 / self.sig_abs * ODsumcond  # Condensed atom number
        Ntherm = self.pix_cal**2 / self.sig_abs * ODsumtherm  # Thermal atom  number
        Ntot = self.pix_cal**2 / self.sig_abs * ODsumtot  # Thermal atom  number

        if print_results:
            print("Total Atom Number: ", int(Ntot))
            print("Condensed Atom Number: ", int(Ncond))
            print("Thermal Atom Number: ", int(Ntherm))

        return [Ntot, Ncond, Ntherm]

    def calculate_temperature(self, bimodalfit_params):
        Gfit_params = [
            bimodalfit_params[0],
            bimodalfit_params[6],
            bimodalfit_params[7],
            bimodalfit_params[1],
            bimodalfit_params[2],
            bimodalfit_params[8],
        ]

        TOF = 0
        TOF = self.input.time_of_flight_ms * 1e-3
        # Time of flight
        tau_x = self.omega_x * TOF  # Dimensionless expansion coefficient
        tau_y = self.omega_y * TOF
        pixcal = 8.71 * 1e-6
        # pixel calibration - input sizes as # of pixels
        mRb = 87 * 1.66 * 1e-27
        # 87Rb mass
        kb = 1.3806 * 1e-23

        sigma_x = Gfit_params[3] * pixcal
        sigma_y = Gfit_params[4] * pixcal

        T_x = (
            (mRb / 2 / kb)
            * (np.square(self.omega_x) * np.square(sigma_x))
            / (1 + np.square(self.omega_x) * np.square(TOF))
        )
        T_y = (
            (mRb / 2 / kb)
            * (np.square(self.omega_y) * np.square(sigma_y))
            / (1 + np.square(self.omega_y) * np.square(TOF))
        )

        T = (
            2 * np.square(tau_x) / (1 + 3 * np.square(tau_x)) * T_x
            + (1 + np.square(tau_y)) / (1 + 3 * np.square(tau_y)) * T_y
        )
        T_nK = T * 1e9

        return T_nK



class JobReadError(Exception):
    def __init__(self, message):
        super().__init__(message)


class AuthorizationError(Exception):
    def __init__(self, message=None):
        super().__init__(message)


class ValidationError(Exception):
    def __init__(self, message):
        super().__init__(message)


class JobWriteError(Exception):
    pass


class DatabaseError(Exception):
    def __init__(self, message):
        super().__init__(message)


class ParameterError(Exception):
    def __init__(self, message):
        super().__init__(message)


class JobError(Exception):
    pass


class JobPlotFitError(JobError):
    def __init__(self):
        super().__init__(
            "PLOT FIT RESULTS: failed to generate model from provided parameters"
        )


class JobPlotFitMismatchError(JobError):
    def __init__(self):
        super().__init__("PLOT FIT RESULTS: mismatched parameters and model type")


class VersionWarning(Warning):
    def __init__(self, message):
        super().__init__(message)



import json
import os
import sys
import time
import warnings
from importlib.metadata import version

import jwt
import requests
import semver
from bert_schemas import job as job_schema
from pydantic import ValidationError

from oqtant.schemas.job import AlbertJob
from oqtant.settings import Settings
from oqtant.util import exceptions as api_exceptions

settings = Settings()


class OqtantClient:
    """Python class for interacting with Albert
    This class contains tools for:
        - Accessing all of the functionality of the Albert Web App (https://albert-dev.coldquanta.com)
            - BARRIER (Barrier Manipulator) jobs
            - BEC (Ultracold Matter) jobs
        - Building parameterized (i.e. optimization) experiments using AlbertJobs
        - Submitting and retrieving AlbertJob results
    How Oqtant works:
        1.) Construct a single or list of AlbertJobs using 'generate_albert_job()'
        2.) Run the single or list of AlbertJobs on the Albert hardware using 'run_jobs()'
            - There is a limit of 30 AlbertJobs per use of 'run_jobs()'
        3.) As AlbertJobs are running, the results are automatically stored in 'active_jobs'
            - The AlbertJobs stored in 'active_jobs' are available until the python session ends
        4.) If you choose to not track the status of AlbertJobs with 'run_jobs()' you can see the status
            of your session's active AlbertJobs with 'see_active_jobs()'
        5.) To operate on jobs submitted in a previous session you can load them into your 'active_jobs'
            by using either 'load_job_from_id()' or 'load_job_from_file()'
        6.) To analyze AlbertJob objects and use Albert's job analysis library reference the AlbertJob
            class documentation located in 'oqtant/job.py'
    Need help? Found a bug? Contact albert@infleqtion.com for support. Thank you!
    """

    def __init__(self, *, settings, token, debug: bool = False):
        self.base_url: str = settings.base_url
        self.active_jobs: dict[str, AlbertJob] = {}
        self.token: str = token
        self.max_ind_var: int = settings.max_ind_var
        self.max_job_batch_size: int = settings.max_job_batch_size
        self.debug: bool = debug
        self.version = version("oqtant")

        if not self.debug:
            sys.tracebacklimit = 0

    def __get_headers(self) -> dict:
        """Generate headers for use in calls to the REST API with requests
        Returns:
            dict: a dict of header information
        """
        return {
            "Authorization": f"Bearer {self.token}",
            "X-Client-Version": version("oqtant"),
        }

    def get_job(self, job_id: str, run: int = 1) -> AlbertJob:
        """Gets an AlbertJob from the Albert REST API. This will always be a targeted query
           for a specific run. If the run is omitted then this will always return the first
           run of the job. Will return results for any job regardless of it's status.
        Args:
            job_id (str): this is the external_id of the job to fetch
            run (int): the run to target, this defaults to the first run if omitted
        Returns:
            AlbertJob: an AlbertJob instance with the values of the job queried
        """
        request_url = f"{self.base_url}/{job_id}"
        params = {"run": run}
        response = requests.get(
            url=request_url,
            params=params,
            headers=self.__get_headers(),
            timeout=(5, 30),
        )
        if response.status_code in [401, 403]:
            raise api_exceptions.AuthorizationError
        response.raise_for_status()
        job_data = response.json()
        try:
            job = AlbertJob(**job_data)
        except (KeyError, ValidationError) as err:
            raise api_exceptions.ValidationError(f"Failed to get job {job_id}: {err}")
        return job

    def get_job_inputs_without_output(
        self, job_id: str, run: int | None = None, include_notes: bool = False
    ) -> dict:
        """Gets an AlbertJob from the Albert REST API. This can return all runs within a job
           or a single run based on whether a run value is provided. The AlbertJobs returned
           will be converted to dictionaries and will not have any output data, even if
           they are complete. This is useful for taking an existing job and creating a new one
           based on it's input data.
        Args:
           job_id (str): this is the external_id of the job to fetch
           run (Union[int, None]): optional argument if caller wishes to only has a single run returned
           include_notes (bool): optional argument if caller wishes to include any notes associated
              with AlbertJob inputs. Defaults to False is not provided
        Returns:
           dict: a dict representation of an AlbertJob instance
        """
        request_url = f"{self.base_url}/{job_id}"
        params = {"exclude_input_output": True}
        if run:
            params["run"] = run
        response = requests.get(
            url=request_url,
            params=params,
            headers=self.__get_headers(),
            timeout=(5, 30),
        )
        if response.status_code in [401, 403]:
            raise api_exceptions.AuthorizationError
        response.raise_for_status()
        job_data = response.json()
        try:
            job = AlbertJob(**job_data)
        except (KeyError, ValidationError) as err:
            raise api_exceptions.ValidationError(f"Failed to get job {job_id}: {err}")
        if not include_notes:
            job.inputs[0].notes = ""
        job = job.dict()
        job.pop("input_count")
        return job

    def generate_albert_job(self, *, job: dict) -> AlbertJob:
        """Generates an instance of AlbertJob from the provided dictionary that contains the
           job details and input. Will validate the values and raise an informative error if
           any violations are found.
        Args:
           job (dict): dictionary containing job details and input
        Returns:
           AlbertJob: an AlbertJob instance containing the details and input from the provided
              dictionary
        """
        try:
            albert_job = AlbertJob(**job)
        except ValidationError as err:
            raise api_exceptions.ValidationError(f"Failed to generate AlbertJob: {err}")
        return albert_job

    def submit_job(self, *, job: AlbertJob) -> dict:
        """Submits a single AlbertJob to the Albert REST API. Upon successful submission this
           function will return a dictionary containing the external_id of the job and it's
           position in the queue.
        Args:
           job (AlbertJob): the AlbertJob instance to submit for processing
        Returns:
           dict: dictionary containing the external_id of the job and it's queue position
        """
        if not isinstance(job, AlbertJob):
            try:
                job = AlbertJob(**job)
            except (TypeError, AttributeError, ValidationError) as err:
                raise api_exceptions.ValidationError(f"Job is invalid: {err}")
        data = {
            "name": job.name,
            "job_type": job.job_type,
            "input_count": len(job.inputs),
            "inputs": [input.dict() for input in job.inputs],
        }
        response = requests.post(
            url=self.base_url,
            json=data,
            headers=self.__get_headers(),
            timeout=(5, 30),
        )
        if response.status_code in [401, 403]:
            raise api_exceptions.AuthorizationError
        response.raise_for_status()
        response_data = response.json()
        return response_data

    def run_jobs(
        self,
        job_list: list[AlbertJob],
        track_status: bool = False,
        write: bool = False,
        filename: str | list[str] = "",
    ) -> list[str]:
        """Submits a list of AlbertJobs to the Albert REST API. This function provides some
           optional functionality to alter how it behaves. Providing it with an argument of
           track_status=True will make it wait and poll the Albert REST API until all jobs
           in the list have completed. The track_status functionality outputs each jobs
           current status as it is polling and opens up the ability to use the other optional
           arguments write and filename. The write and filename arguments enable the ability
           to have the results of each completed job written to a file. The value of filename
           is optional and if not provided will cause the files to be created using the
           external_id of each job. If running more than one job and using the filename
           argument it is required that the number of jobs in job_list match the number of
           values in filename.
        Args:
           job_list (list[AlbertJob]): the list of AlbertJob instances to submit for processing
           track_status (bool): optional argument to tell this function to either return
             immediately or wait and poll until all jobs have completed
           write (bool): optional argument to tell this function to write the results of each
             job to file when complete
           filename (Union[str, list[str]]): optional argument to be used in conjunction with the
             write argument. allows the caller to customize the name(s) of the files being created
        Returns:
           list[str]: list of the external_id(s) returned for each submitted job in job_list
        """
        if len(job_list) > self.max_job_batch_size:
            raise AttributeError(
                f"Maximum number of jobs submitted per run is {self.max_job_batch_size}."
            )
        pending_jobs = []
        submitted_jobs = []
        for job in job_list:
            response = self.submit_job(job=job)
            external_id = response["job_id"]
            queue_position = response["queue_position"]
            est_time = response["est_time"]

            job.external_id = external_id
            self.active_jobs[external_id] = job

            pending_jobs.append(external_id)
            submitted_jobs.append(external_id)
            print(
                f"submitted {job.name} ID: {job.external_id} queue_position: {queue_position} \
                    est. time: {est_time} minutes"
            )
        print("Jobs submitted: \n")
        if track_status:
            self.track_jobs(pending_jobs=pending_jobs, filename=filename, write=write)
        return submitted_jobs

    def search_jobs(
        self,
        *,
        job_type: job_schema.JobType | None = None,
        name: job_schema.JobName | None = None,
        submit_start: str | None = None,
        submit_end: str | None = None,
        notes: str | None = None,
    ) -> list[dict]:
        """Submits a query to the REST API to search for jobs that match the provided criteria.
           The search results will be limited to jobs that meet your Albert account access.
        Args:
           job_type (job_schema.JobType): the type of the jobs to search for
           name (job_schema.JobName): the name of the job to search for
           submit_start (str): the earliest submit date of the jobs to search for
           submit_start (str): the latest submit date of the jobs to search for
           notes (str): the notes of the jobs to search for
        Returns:
           list[dict]: a list of jobs matching the provided search criteria
        """
        request_url = f"{self.base_url}/"
        params = {}
        for param in ["job_type", "name", "submit_start", "submit_end", "notes"]:
            if locals()[param] is not None:
                params[param] = locals()[param]

        response = requests.get(
            url=request_url,
            params=params,
            headers=self.__get_headers(),
            timeout=(5, 30),
        )
        if response.status_code in [401, 403]:
            raise api_exceptions.AuthorizationError
        response.raise_for_status()
        job_data = response.json()
        return job_data["items"]

    def track_jobs(
        self,
        *,
        pending_jobs: list[str],
        filename: str | list = "",
        write: bool = False,
    ) -> None:
        """Polls the Albert REST API with a list of job external_ids and waits until all of them have
           completed. Will output each job's status while it is polling and will output a message when
           all jobs have completed. This function provides some optional functionality to alter how it
           behaves. Providing it with an argument of write will have it write the results of each
           completed job to a file. There is an additional argument that can be used with write called
           filename. The value of filename is optional and if not provided will cause the files to be
           created using the external_id of each job. If tracking more than one job and using the
           filename argument it is required that the number of jobs in job_list match the number of
           values in filename.
        Args:
           pending_jobs (list[str]): list of job external_ids to track
           write (bool): optional argument to tell this function to write the results of each job to
             file when complete
           filename (Union[str, list[str]]): optional argument to be used in conjunction with the write
             argument. allows the caller to customize the name(s) of the files being created
        """
        if write:
            if not filename:
                job_filenames = [f"{external_id}.txt" for external_id in pending_jobs]
            elif isinstance(filename, str):
                job_filenames = [filename]
            else:
                job_filenames = filename
            if len(job_filenames) != len(pending_jobs):
                raise AttributeError(
                    "Write filename list length does not match number of jobs."
                )

        status = None
        while pending_jobs:
            for pending_job in pending_jobs:
                job = self.get_job(job_id=pending_job)
                self.active_jobs[pending_job] = job
                if job.status != status:
                    print(f"job {pending_job} is in status {job.status}")
                    status = job.status
                if job.status == job_schema.JobStatus.COMPLETE:
                    print(f"job complete: {pending_job}")
                    pending_jobs.remove(pending_job)
                    if write:
                        self._write_job_to_file(
                            self.active_jobs[pending_job], job_filenames.pop(0)
                        )
                if job.status in [
                    job_schema.JobStatus.INCOMPLETE,
                    job_schema.JobStatus.FAILED,
                ]:
                    pending_jobs.remove(pending_job)
                    if write:
                        job_filenames.pop(0)
                time.sleep(2)
        print("all jobs complete")

    def load_job_from_id_list(self, job_id_list: list[str]) -> None:
        """Loads AlbertJobs from the Albert REST API into the current active_jobs list using a list
           of job external_ids. The results of the jobs loaded by this function are limited to their
           first run.
        Args:
           job_id_list (list[str]): list of job external_ids to load
        """
        for job_id in job_id_list:
            self.load_job_from_id(job_id)

    def load_job_from_id(self, job_id: str, run: int = 1) -> None:
        """Loads an AlbertJob from the Albert REST API into the current active_jobs list using a job
           external_id. The results of the jobs loaded by this function can be targeted to a specific
           run if there are multiple.
        Args:
           job_id (str): the external_id of the job to load
           run (int): optional argument to target a specific job run
        """
        try:
            job = self.get_job(job_id=job_id, run=run)
            self.active_jobs[job_id] = job
            print(f"Loaded job: {job.name} {job_id}")
        except Exception as err:
            raise api_exceptions.ValidationError(
                f"Failed to fetch job {job_id}: {err}. Please contact ColdQuanta if error persists"
            )

    def load_job_from_file_list(self, file_list: list[str]) -> None:
        """Loads AlbertJobs from the Albert REST API into the current active_jobs list using a list
           of filenames containing AlbertJob info. The results of the jobs loaded by this function are
           limited to their first run.
        Args:
           file_list (list[str]): list of filenames containing AlbertJob information
        """
        for f in file_list:
            self.load_job_from_file(f)

    def load_job_from_file(self, file: str) -> None:
        """Loads an AlbertJob from the Albert REST API into the current active_jobs list using a file
           containing AlbertJob info. The results of the jobs loaded by this function are limited to
           their first run.
        Args:
           file_list (list[str]): list of filenames containing AlbertJob information
        """
        try:
            with open(file) as json_file:
                data = json.load(json_file)
                self.load_job_from_id(data["external_id"])
        except (FileNotFoundError, Exception) as err:
            raise api_exceptions.JobReadError(f"Failed to load job from {file}: {err}")

    def see_active_jobs(self, refresh: bool = True) -> None:
        """Utility function to print out the current contents of the active_jobs list. The optional
           argument of refresh tells the function whether it should refresh the data of pending or
           running jobs stored in active_jobs before printing out the results. Refreshing also
           updates the data in active_jobs so if jobs were submitted but not tracked this is a way
           to check on their status.
        Args:
           refresh (bool): optional argument to refresh the data of jobs in active_jobs
        """
        if refresh:
            for external_id, job in self.active_jobs.items():
                if job.status in [
                    job_schema.JobStatus.PENDING,
                    job_schema.JobStatus.RUNNING,
                ]:
                    refreshed_job = self.get_job(
                        job_id=external_id, run=job.inputs[0].run
                    )
                    self.active_jobs[external_id] = refreshed_job
        print("ACTIVE JOBS")
        print("NAME\t\tSTATUS\t\tTIME SUBMIT\t\tID")
        print("_" * 50)
        for job_id, job in self.active_jobs.items():
            print(f"{job.name}\t\t{job.status}\t\t{job.time_submit}\t\t{job_id}")

    def _write_job_to_file(self, job: AlbertJob, filepath: str) -> None:
        """Utility function to write an AlbertJob instance to a file. Requires the full filepath
           including the name of the file the write. Files that already exist cannot be
           overwritten.
        Args:
           job (AlbertJob): the AlbertJob instance to write to file
           filepath (str): the full path to the file to write, including the name of the file
        """
        if os.path.exists(filepath):
            raise api_exceptions.JobWriteError(
                "File already exists. Please choose a unique filename."
            )

        try:
            with open(filepath, "w") as f:
                f.write(str(job.json()))
            print(f"Wrote job {job.external_id} to file {filepath}")
        except (FileNotFoundError, Exception) as err:
            print(f"Failed to write job {job.external_id} to {filepath}: {err}")

    def get_job_limits(self) -> dict:
        """Utility method to get job limits from the REST API
        Returns:
            dict: dictionary of job limits
        """
        try:
            token_data = jwt.decode(
                self.token, key=None, options={"verify_signature": False}
            )
            external_user_id = token_data["sub"]
        except Exception:
            raise api_exceptions.ValidationError(
                "Unable to decode JWT token. Please contact ColdQuanta."
            )

        url = f"{self.base_url.replace('jobs', 'users')}/{external_user_id}/job_limits"
        response = requests.get(
            url=url,
            headers=self.__get_headers(),
            timeout=(5, 30),
        )
        if response.status_code in [401, 403]:
            raise api_exceptions.AuthorizationError("Unauthorized")
        response.raise_for_status()
        job_limits = response.json()
        return job_limits


def version_check(client_version: str) -> None:
    """Compares the given current Oqtant version with the version currently on pypi, and raises a warning if it is older.

    Args:
        client_version (str): the client semver version number
    """
    resp = requests.get("https://pypi.org/pypi/oqtant/json", timeout=5)
    if resp.status_code == 200:
        current_version = resp.json()["info"]["version"]
        if semver.compare(client_version, current_version) < 0:
            warnings.warn(
                f"Please upgrade to Oqtant version {current_version}. You are currently using version {client_version}."
            )


def get_oqtant_client(token: str) -> OqtantClient:
    """A utility function to create a new OqtantClient instance.
    Args:
        token (str): the auth0 token required for interacting with the REST API
    Returns:
        OqtantClient: authenticated instance of OqtantClient
    """

    client = OqtantClient(settings=settings, token=token)
    version_check(client.version)
    return client
